from typing import Any, Type, Union

from ..dom import DOMInfo, DOMDict

from .SchemaObject import SchemaObject
from ..base_schema import Schema
from .resolve_arg_to_type import resolve_arg_to_schema


class SchemaDict(SchemaObject):
    """
    A schema specifying an object with dynamic properties (corresponding to a Python dict)

    :param items: The permitted data type or schema for the properties of this object. Must
                  be one of:
                  A primitive Python type (str, int, bool, float)
                  A subclass of `UserObject`
                  An instance of `Schema`
    """

    def __init__(
            self,
            items: Union[Type, Schema]
    ) -> None:
        super().__init__(
            additional_properties=resolve_arg_to_schema(items)
        )

    def __call__(
            self,
            value: Any,
            dom_info: DOMInfo = None
    ) -> Any:
        return DOMDict(
            value,
            dom_info,
            item_type=self.additional_properties
        )



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Dict, Tuple, Optional

import jsonschema
from jsonschema.validators import validator_for

from ..exceptions import ValidationError


class Schema(ABC):
    """
    Abstract base class for JSON schemas. Objects of type `Schema` can be used to generate
    jsonschema-compatible dictionaries, and to validate potential input data against those
    schemas.

    Objects of type `Schema` are also callable, and when called will create DOM objects or
    primitive Python object containing the data that is supplied to them.
    """

    def __call__(
            self,
            value: Any,
            dom_info: Tuple = None
    ) -> Any:
        """
        Return either a DOM object or primitive Python object containing the data
        supplied in `value`, if `value` is a valid instance of this schema.

        :param value:    The raw value for which to create an object
        :param dom_info: An optional tuple containing DOM information for the object, if relevant
        :return:         A DOM object or primitive Python object containing the data in `value`
        """
        if self.is_valid(value):
            return value
        else:
            raise ValidationError(
                f"The supplied value does not conform to this schema: {value}"
            )

    @property
    def schema_ref_name(self) -> Optional[str]:
        """
        A unique reference name to use when this schema is referred to by other schemas.
        If this returns a string, references to this schema will use the $ref keyword without
        replicating the full schema.
        If this property returns None, the full contents of the schema will be used.

        :return: A string with a unique reference name if defined, else None.
        """
        return None

    @property
    def referenced_schemas(self) -> Dict[str, Schema]:
        """
        A dict of named schemas (i.e. schemas with a defined `schema_ref_name`) that are
        referenced in this schema, including itself if applicable.

        :return: A dict of `Schema` objects indexed by their `schema_ref_name`.
        """
        return {}

    @property
    @abstractmethod
    def jsonschema_definition(self) -> Dict[str, Any]:
        """
        The underlying jsonschema-compatible schema definition for this schema.

        :return: A jsonschema-compatible dictionary.
        """
        pass

    @property
    def jsonschema_ref_schema(self) -> Dict[str, Any]:
        """
        The jsonschema definition to use when referring to this schema in another schema.
        If `schema_ref_name` is defined, this will be a reference using the "$ref" keyword.
        If `schema_ref_name` is None, the raw definition of the schema will be used.

        :return: A jsonschema-compatible dictionary.
        """
        if self.schema_ref_name:
            return {"$ref": f"#/definitions/{self.schema_ref_name}"}
        else:
            return self.jsonschema_definition

    @property
    def jsonschema_full_schema(self) -> Dict[str, Any]:
        """
        The jsonschema definition to use when using this schema as a standalone schema.

        :return: A jsonschema-compatible dictionary.
        """
        output_schema = {
            "$schema": "http://json-schema.org/draft-07/schema#",
            "definitions": {
                k: v.jsonschema_definition
                for k, v in self.referenced_schemas.items()
            }
        }
        output_schema.update(self.jsonschema_ref_schema)
        return output_schema

    def validate(self, value: Any) -> None:
        """
        Determine whether a given object conforms to this schema, and throw an error if not.

        :param value: An object to test for validity against this schema
        """
        jsonschema.validate(value, self.jsonschema_full_schema)

    def is_valid(self, value: Any) -> bool:
        """
        Determine whether a given object conforms to this schema.

        :param value: An object to test for validity against this schema
        :return:      True if the object is valid, otherwise False
        """
        return validator_for(self.jsonschema_full_schema)(self.jsonschema_full_schema).is_valid(value)



from .__version__ import __version__
from .exceptions import ValidationError
from .dom import document, parent, key, schema
from . import dom
from .base_schema import Schema, SchemaType, SchemaNone, SchemaPrimitive, SchemaAnything, SchemaConst
from .object_schema import SchemaArray, SchemaDict, SchemaAnyOf, SchemaObject
from .user_objects import UserProperty, UserObject
from .mixins import ReadsJSON, ReadsYAML, RegistersSubclasses




from typing import Optional, Dict, Type, Any, Union, Set

from ..dom import DOMInfo
from ..base_schema import SchemaType
from ..base_schema import Schema


class SchemaObject(SchemaType):
    """
    A schema specifying an object with named properties.

    :param properties:            A dictionary of `Schema` objects defining the expected
                                  names and types of this object's properties.
    :param additional_properties: Defines whether this object permits additional
                                  dynamically-named properties. Can be True or False, or
                                  can be set to a specific `Schema` to restrict the permitted
                                  types of any additional properties.
    :param object_type:           A custom object type to use when creating object instances
                                  from this schema.
    :param schema_ref_name:       An optional unique reference name to use when this schema
                                  is referred to by other schemas.
    """

    type_name: str = "object"
    properties: Optional[Dict[str, Schema]] = None
    required: Set[str] = None
    additional_properties: Union[bool, Schema] = False
    schema_ref_name: Optional[str] = None

    def __init__(
            self,
            properties: Optional[Dict[str, Schema]] = None,
            required: Set[str] = None,
            additional_properties: Union[bool, Schema] = False,
            object_type: Type = dict,
            schema_ref_name: Optional[str] = None
    ) -> None:
        self.properties = properties or {}
        self.required = required or set()
        self.additional_properties = additional_properties
        self.object_type = object_type
        self.schema_ref_name = schema_ref_name

    def __call__(
            self,
            value: Any,
            dom_info: DOMInfo = None
    ) -> Any:
        return self.object_type(value, dom_info)

    @property
    def referenced_schemas(self) -> Dict[str, Schema]:
        referenced_schemas = {}
        if isinstance(self.additional_properties, Schema):
            referenced_schemas.update(self.additional_properties.referenced_schemas)
        for schema in self.properties.values():
            referenced_schemas.update(schema.referenced_schemas)
        if isinstance(self.schema_ref_name, str):
            referenced_schemas[self.schema_ref_name] = self
        return referenced_schemas

    @property
    def jsonschema_definition(self) -> Dict[str, Any]:
        return {
            **super().jsonschema_definition,
            "properties": {
                k: v.jsonschema_ref_schema
                for k, v in self.properties.items()
            },
            "required": list(sorted(self.required)),
            "additionalProperties": (
                self.additional_properties.jsonschema_ref_schema
                if isinstance(self.additional_properties, Schema)
                else self.additional_properties
            )
        }





from typing import Any, Dict

from .Schema import Schema


class SchemaConst(Schema):
    """
    A schema requiring a string constant.
    """

    value: str = None

    def __init__(
            self,
            value: str
    ) -> None:
        self.value = value

    @property
    def jsonschema_definition(self) -> Dict[str, Any]:
        return {
            "const": self.value
        }



from __future__ import annotations

from typing import Any, Type, Iterator, Union, Mapping

import inspect

from ..mixins import RegistersSubclasses
from ..base_schema import Schema
from ..object_schema import SchemaObject, SchemaAnyOf
from ..dom import (
    DOMObject,
    DOMProperties,
    DOMInfo
)

from .UserProperty import UserProperty


class UserProperties(DOMProperties):
    """
    A container for property information for a :class:`.UserObject` subclass.
    """

    def __init__(
            self,
            user_class: Type[UserObject],
            additional_properties: Union[bool, Schema] = False
    ):
        """
        :param user_class:            The subclass of :class:`.UserObject` to extract properties from.
                                      Properties will be extracted from the class's
                                      :class:`~wysdom.user_objects.UserProperty` descriptors.
        :param additional_properties: Defines whether a :class:`.DOMObject` permits additional
                                      dynamically-named properties. Can be True or False, or
                                      can be set to a specific :class:`~wysdom.base_schema.Schema`
                                      to restrict the permitted types of any additional properties.
        """
        self._user_class = user_class
        properties = {}
        required = set()
        for superclass in reversed(list(self._schema_superclasses())):
            for k, v in superclass.__dict__.items():
                if isinstance(v, UserProperty):
                    if not v.name:
                        v.name = k
                    properties[v.name] = v.schema_type
                    if not v.optional:
                        required.add(v.name)
        super().__init__(properties, required, additional_properties)

    def _schema_superclasses(self) -> Iterator[Type[UserObject]]:
        for superclass in inspect.getmro(self._user_class):
            if (
                    issubclass(superclass, UserObject)
                    and superclass is not UserObject
            ):
                yield superclass


class UserObject(DOMObject):
    """
    Base class for user-defined DOM objects.

    :param value:         A dict-like object to populate the underlying data
                          object's keys. May be provided alone or in conjunction
                          with keyword arguments.

    :param json_dom_info: A named tuple with parameters (element, document, parent,
                          element_key) which specify this object's position in a
                          larger document object model. This is only required if you
                          need to set these values when creating an object manually:
                          when objects are created as part of a DOM specification,
                          these values are populated automatically.

    :param kwargs:        Keyword arguments to be used to populate the underlying
                          data object's keys. May be provided alone or in
                          conjunction with `value`.
    """

    __json_schema_properties__: UserProperties = None

    def __init_subclass__(
            cls,
            *args: Any,
            additional_properties: Union[bool, Schema] = False,
            **kwargs: Any
    ) -> None:
        cls.__json_schema_properties__ = UserProperties(cls)
        super().__init_subclass__(*args, **kwargs)

    def __init__(
            self,
            value: Mapping[str, Any] = None,
            json_dom_info: DOMInfo = None,
            **kwargs: Any
    ) -> None:
        if json_dom_info:
            if not isinstance(json_dom_info, DOMInfo):
                raise TypeError(
                    "The name json_dom_info is a reserved parameter of "
                    "UserObject and must be of type "
                    f"DOMInfo, not {type(json_dom_info)}"
                )
        super().__init__({**(value or {}), **kwargs}, json_dom_info)

    @classmethod
    def __json_schema__(cls) -> Schema:
        has_subclasses = False
        if issubclass(cls, RegistersSubclasses):
            if cls.registered_subclasses():
                has_subclasses = True
        if has_subclasses:
            return SchemaAnyOf(
                (
                    subclass.__json_schema__()
                    for subclass_list in cls.registered_subclasses().values()
                    for subclass in subclass_list
                    if issubclass(subclass, UserObject)
                    and not isinstance(subclass.__json_schema__(), SchemaAnyOf)
                ),
                schema_ref_name=f"{cls.__module__}.{cls.__name__}"
            )
        else:
            return SchemaObject(
                properties=cls.__json_schema_properties__.properties,
                required=cls.__json_schema_properties__.required,
                additional_properties=cls.__json_schema_properties__.additional_properties,
                object_type=cls,
                schema_ref_name=f"{cls.__module__}.{cls.__name__}"
            )



from __future__ import annotations

from typing import Any, Union, TextIO

import yaml

from ..dom import DOMObject


class ReadsYAML(DOMObject):
    """
    Adds YAML reading and writing functionality to a DOMObject.
    """

    def to_yaml(self, **kwargs: Any) -> str:
        """
        Serialize the DOM object to YAML.

        :param kwargs: Optional keyword arguments to pass to PyYAML's safe_dump method.
                       See parameters for Dumper in https://pyyaml.org/wiki/PyYAMLDocumentation
        :return:       The DOM object, serialized as a YAML string
        """
        return yaml.safe_dump(self.to_builtin(), **kwargs)

    @classmethod
    def from_yaml(
            cls,
            yaml_string: Union[str, TextIO]
    ) -> ReadsYAML:
        """
        Create a new DOM object by from a YAML string.

        :param yaml_string: YAML string to read
        :return:            New DOM object instance
        """
        return cls(yaml.safe_load(yaml_string))

    @classmethod
    def from_yaml_file(cls, filename: str) -> ReadsYAML:
        """
        Create a new DOM object from a file on disk.

        :param filename: File path on disk of YAML file
        :return:         New DOM object instance
        """
        with open(filename, "r") as stream:
            return cls.from_yaml(stream)




from __future__ import annotations

from abc import ABC

from typing import Type, Optional, Any, Dict, List


class RegistersSubclasses(ABC):
    """
    A mixin class that allows subclasses to be registered in the
    main class by a registered name, using the `register_as`
    parameter in the subclass declaration.
    """

    registered_name = None
    __registered_subclasses__: Optional[Dict[str, RegisteredSubclassList]] = None

    def __init_subclass__(
        cls,
        register_as: Optional[str] = None,
        **kwargs: Any
    ) -> None:
        super().__init_subclass__(**kwargs)

        if cls.__registered_subclasses__ is None and RegistersSubclasses in cls.__bases__:
            cls.__registered_subclasses__ = {}

        name = register_as or f"{cls.__module__}.{cls.__name__}"
        if not isinstance(name, str):
            raise TypeError(
                f"Parameter register_as must be a string, not {type(register_as)}")

        cls.__registered_subclasses__.setdefault(name, [])
        if cls not in cls.__registered_subclasses__[name]:
            cls.__registered_subclasses__[name].append(cls)
        cls.registered_name = name

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

    @classmethod
    def registered_subclasses(cls) -> Dict[str, RegisteredSubclassList]:
        """
        Return all of the registered subclasses in this class's namespace.

        :return: A dictionary of subclasses, indexed by registered name.
        """
        return {
            name: cls.registered_subclasses_by_name(name)
            for name in cls.__registered_subclasses__.keys()
            if len(cls.registered_subclasses_by_name(name)) > 0
        }

    @classmethod
    def registered_subclasses_by_name(cls, name) -> RegisteredSubclassList:
        """
        Return all of the registered subclasses for a given name that are a
        proper subclass of this class.

        :param name: The registered name of the subclass, as defined in
                     `register_as` when the class was declared.
        :return:     A list of subclasses of the class from which this method was called.
        """
        return [
            subclass
            for subclass in cls.__registered_subclasses__.get(name, [])
            if issubclass(subclass, cls)
            and subclass is not cls
        ]

    @classmethod
    def registered_subclass(
            cls,
            name: str,
            return_common_superclass: bool = True
    ) -> Type[RegistersSubclasses]:
        """
        Return a registered subclass by name.

        :param name:                     The registered name of the subclass, as defined in
                                         `register_as` when the class was declared.
        :param return_common_superclass: Disambiguate multiple valid subclasses by returning
                                         the common superclass among them, if one exists.
                                         Defaults to True.
        :return:                         A subclass of the class from which this method was called.
        :raises KeyError:                If no matching subclass is found or if multiple ambiguous
                                         subclasses are found.
        """
        matched_subclasses = cls.registered_subclasses_by_name(name)
        if len(matched_subclasses) == 0:
            raise KeyError(
                f"The key '{name}' matches no proper subclasses of {cls}.")
        elif len(matched_subclasses) > 1:
            if return_common_superclass:
                for matched_subclass in matched_subclasses:
                    if all(
                            issubclass(other_subclass, matched_subclass)
                            for other_subclass in matched_subclasses
                    ):
                        return matched_subclass
            raise KeyError(
                f"""
                The key '{name}' is ambiguous as it matches multiple proper subclasses of {cls}:
                {matched_subclasses}
                """)
        return matched_subclasses[0]

    @classmethod
    def registered_subclass_instance(
        cls,
        name: str,
        *args: Any,
        **kwargs: Any
    ) -> RegistersSubclasses:
        """
        Create a new instance of a registered subclass by name.

        :param name:   The registered name of the subclass, as defined in
                       `register_as` when the class was declared.
        :param args:   Positional arguments to pass to the subclass.
        :param kwargs: Keyword arguments to pass to the subclass.
        :return:       An instance of a subclass of the class from which this method was called.
        """
        return cls.registered_subclass(name)(*args, **kwargs)


RegisteredSubclassList = List[Type[RegistersSubclasses]]



from __future__ import annotations

from typing import Generic, TypeVar, Optional, Any, Dict

from collections.abc import Mapping

from ..base_schema import Schema, SchemaAnything
from .DOMElement import DOMElement
from .DOMObject import DOMObject
from . import DOMInfo
from .DOMProperties import DOMProperties

T_co = TypeVar('T_co')


class DOMDict(DOMObject, Generic[T_co]):
    """
    An object with dynamic properties (corresponding to a Python dict).
    """

    def __init__(
            self,
            value: Optional[Mapping[str, Any]] = None,
            json_dom_info: Optional[DOMInfo] = None,
            item_type: Optional[Schema] = None
    ) -> None:
        """
        :param value:         A dict (or any :class:`collections.abc.Mapping`) containing the data to populate this
                              object's properties.
        :param json_dom_info: A :class:`~wysdom.dom.DOMInfo` named tuple containing information about this object's
                              position in the DOM.
        :param item_type:     A :class:`~wysdom.Schema` object specifying what constitutes a valid property
                              of this object.
        """
        self.__json_schema_properties__ = DOMProperties(
            additional_properties=(item_type or SchemaAnything())
        )
        super().__init__(
            value or {},
            json_dom_info
        )

    def __getitem__(self, key: str) -> T_co:
        return super().__getitem__(key)

    def __deepcopy__(self, memo: Dict[int, DOMElement]) -> DOMDict:
        cls = self.__class__
        result = cls(
            value=self.to_builtin(),
            json_dom_info=self.__json_dom_info__,
            _item_type=self.__json_schema_properties__.additional_properties
        )
        memo[id(self)] = result
        return result



from __future__ import annotations

from typing import Any, Optional, Iterator, Dict

from collections.abc import Mapping, MutableMapping

from ..exceptions import ValidationError
from ..base_schema import SchemaAnything

from .DOMElement import DOMElement
from . import DOMInfo
from .DOMProperties import DOMProperties
from .functions import document, schema


class DOMObject(DOMElement, MutableMapping):
    """
    An object with named properties.
    """

    __json_schema_properties__: DOMProperties = None
    __json_element_data__: Dict[str, Optional[DOMElement]] = None

    def __init__(
            self,
            value: Mapping[str, Any] = None,
            json_dom_info: DOMInfo = None
    ) -> None:
        """
        :param value:         A dict (or any :class:`collections.abc.Mapping`) containing the data to populate this
                              object's properties.
        :param json_dom_info: A :class:`~wysdom.dom.DOMInfo` named tuple containing information about this object's
                              position in the DOM.
        """
        if value and not isinstance(value, Mapping):
            raise ValidationError(
                f"Cannot validate input. Object is not a mapping: {value}"
            )
        schema(self).validate(value)
        super().__init__(None, json_dom_info)
        self.__json_element_data__ = {}
        try:
            for key, value in value.items():
                self[key] = value
        except KeyError as e:
            raise ValidationError(str(e))

    def __getitem__(self, key: str) -> Optional[DOMElement]:
        return self.__json_element_data__[key]

    def __setitem__(self, key: str, value: Optional[DOMElement]) -> None:
        if value is None:
            if key in self.__json_schema_properties__.required:
                raise ValueError(
                    f"The property '{key}' is not optional and cannot be None."
                )
            self.__json_element_data__[key] = None
        else:
            item_class = self.__json_schema_properties__.properties.get(
                key, self.__json_schema_properties__.additional_properties)
            if item_class is True:
                item_class = SchemaAnything()
            if not item_class:
                raise KeyError(
                    f"No property named '{key}' exists, and "
                    "additional properties are not allowed."
                )
            self.__json_element_data__[key] = item_class(
                value,
                DOMInfo(
                    document=document(self),
                    parent=self,
                    element_key=key
                )
            )

    def __delitem__(self, key: str) -> None:
        del self.__json_element_data__[key]

    def __len__(self) -> int:
        return len(self.__json_element_data__)

    def __iter__(self) -> Iterator[str]:
        return iter(self.__json_element_data__)

    def __repr__(self):
        return f"{self.__class__.__name__}({repr(self.__json_element_data__)})"

    def __str__(self):
        return str(self.__json_element_data__)

    def walk_elements(self) -> Iterator[DOMInfo]:
        yield self.__json_dom_info__
        for key, value in self.items():
            if isinstance(value, DOMElement):
                yield from value.walk_elements()
            else:
                yield DOMInfo(value, document(self), self, key)

    def to_builtin(self) -> Dict[str, Any]:
        """
        Returns the contents of this DOM object as a Python builtin.

        :return: A Python dict containing this object's data
        """
        return {
            k: (
                v.to_builtin()
                if isinstance(v, DOMElement)
                else v
            )
            for k, v in self.items()
        }

    def __copy__(self) -> DOMObject:
        cls = self.__class__
        result = cls.__new__(cls)
        super(DOMObject, result).__init__(json_dom_info=self.__json_dom_info__)
        result.__json_element_data__ = dict(self.__json_element_data__)
        return result

    def __deepcopy__(self, memo: Dict[int, DOMElement]) -> DOMObject:
        cls = self.__class__
        result = cls(
            value=self.to_builtin(),
            json_dom_info=self.__json_dom_info__
        )
        memo[id(self)] = result
        return result



from typing import Type, Any, Union, Optional, Callable

from ..base_schema import Schema
from ..object_schema import resolve_arg_to_schema
from ..dom import DOMObject, DOMInfo, document


class UserProperty(object):
    """
    A data descriptor for creating attributes in user-defined subclasses
    of :class:`~.wysdom.user_objects.UserObject` which are mapped to keys in the underlying
    data object and to the `properties` key in the object's JSON schema.

    :param property_type:    The data type or schema for this property. Must
                             be one of:

                             * A primitive Python type (:class:`str`, :class:`int`,
                               :class:`bool`, :class:`float`)
                             * A subclass of :class:`~.wysdom.user_objects.UserObject`
                             * An instance of :class:`~wysdom.base_schema.Schema`

    :param optional:         Determines whether this property is optional in the underlying
                             data object. If default or default_function are set, this
                             will default to True, otherwise False.

    :param name:             The name of this property in the underlying
                             data object. If not provided, this defaults to
                             the name of the attribute on the :class:`~.wysdom.user_objects.UserObject`
                             instance that owns the property.

    :param default:          A static value which provides a default value
                             for this property. Cannot be set in conjunction
                             with `default_function`.

    :param default_function: A function which provides a default value
                             for this property. The function must have a
                             single positional argument, `self`, which is
                             passed the :class:`~.wysdom.user_objects.UserObject` instance that
                             owns the property. Cannot be set in conjunction
                             with `default`.

    :param persist_defaults: If this property is set to True and a UserProperty has either the
                             `default` or `default_function` property, when the UserProperty returns
                             a default value that value will also be explicitly stored in the underlying
                             data object. This is often desirable behavior if the UserProperty
                             returns another object and your code expects it to return the same
                             object instance each time it is accessed.
    """

    def __init__(
            self,
            property_type: Union[Type, Schema],
            optional: Optional[bool] = None,
            name: Optional[str] = None,
            default: Optional[Any] = None,
            default_function: Optional[Callable] = None,
            persist_defaults: Optional[bool] = None
    ) -> None:
        if default is not None or default_function is not None:
            if default is not None and default_function is not None:
                raise ValueError("Cannot use both default and default_function.")
            if optional is False:
                raise ValueError(
                    "Cannot set optional to False if default or default_function are specified.")
            self.optional = True
        else:
            self.optional = bool(optional)
        self.schema_type = resolve_arg_to_schema(property_type)
        self.name = name
        self.default = default
        self.default_function = default_function
        self.persist_defaults = persist_defaults

    def __get__(
            self,
            instance: DOMObject,
            owner: Type[DOMObject]
    ) -> Any:
        if instance is None:
            raise AttributeError(
                "UserProperty is not valid as a class data descriptor")
        if self.name not in instance:
            if self.default_function:
                default_value = self.default_function(instance)
            else:
                default_value = self.default
            if self.persist_defaults:
                instance[self.name] = default_value
                return instance[self.name]
            elif default_value is None:
                return default_value
            else:
                return self.schema_type(
                    default_value,
                    DOMInfo(
                        document=document(instance),
                        parent=instance,
                        element_key=self.name
                    )
                )
        return instance[self.name]

    def __set__(
            self,
            instance: DOMObject,
            value: Any
    ) -> None:
        instance[self.name] = value



from .SchemaArray import SchemaArray
from .SchemaAnyOf import SchemaAnyOf
from .SchemaObject import SchemaObject
from .SchemaDict import SchemaDict
from .resolve_arg_to_type import resolve_arg_to_schema



from __future__ import annotations

from typing import Any, Iterator, NamedTuple, Optional

from abc import ABC, abstractmethod

from ..base_schema import Schema, SchemaAnything


class DOMInfo(NamedTuple):
    """
    Named tuple containing information about a DOM element's position within the DOM.

    :param element:     The :class:`DOMElement` that this DOMInfo tuple provides information for.
    :param document:    The owning document for a :class:`DOMElement`, if it exists.
    :param parent:      The parent element of a :class:`DOMElement`, if it exists.
    :param element_key: The key of a particular :class:`DOMElement` in its parent element,
                        if it can be referred to by a key (i.e. if it its parent element
                        is a Mapping).
    """

    element: Optional[DOMElement] = None
    document: Optional[DOMElement] = None
    parent: Optional[DOMElement] = None
    element_key: Optional[str] = None


class DOMElement(ABC):
    """
    Abstract base class for any DOM element.
    """

    __json_dom_info__: DOMInfo = None

    @abstractmethod
    def __init__(
            self,
            value: Any = None,
            json_dom_info: DOMInfo = None,
            **kwargs: Any
    ) -> None:
        """
        :param value:         A data structure containing the data to populate this element.
        :param json_dom_info: A :class:`~wysdom.dom.DOMInfo` named tuple containing information about this object's
                              position in the DOM.
        :param kwargs:        Keyword arguments.
        """
        if value is not None:
            raise ValueError(
                "The parameter 'value' must be handled by a non-abstract subclass."
            )
        if json_dom_info:
            self.__json_dom_info__ = DOMInfo(
                element=self,
                document=(
                    self if json_dom_info.document is None
                    else json_dom_info.document
                ),
                parent=json_dom_info.parent,
                element_key=json_dom_info.element_key
            )
        else:
            self.__json_dom_info__ = DOMInfo(
                element=self, document=self)

    @classmethod
    def __json_schema__(cls) -> Schema:
        return SchemaAnything()

    @abstractmethod
    def to_builtin(self) -> Any:
        """
        Returns the contents of this DOM object as a Python builtin. Return type
        varies depending on the specific object type.
        """
        pass

    def walk_elements(self) -> Iterator[DOMInfo]:
        """
        Walk through the full tree structure within this DOM element.
        Returns an iterator of :class:`~wysdom.dom.DOMInfo` tuples in the form
        (element, document, parent element_key).

        :return: An iterator of :class:`~wysdom.dom.DOMInfo` tuples.
        """
        yield self.__json_dom_info__



from .DOMProperties import DOMProperties
from .DOMElement import DOMElement, DOMInfo
from .DOMObject import DOMObject
from .DOMDict import DOMDict
from .DOMList import DOMList
from .functions import document, parent, key, schema



from typing import Any, Dict

from .Schema import Schema


class SchemaAnything(Schema):
    """
    A schema where any valid JSON will be accepted.
    """

    @property
    def jsonschema_definition(self) -> Dict[str, Any]:
        return {}



# -*- coding: utf-8 -*-

import io
import os

from setuptools import setup, find_packages

# Package metadata
# ----------------

NAME = 'wysdom'
DESCRIPTION = (
    'A Python library for building custom document object '
    'models (DOMs) with built-in JSON schema checking'
)
URL = 'https://github.com/jetavator/wysdom'
EMAIL = 'joetaylorconsulting@gmail.com'
AUTHOR = 'Joe Taylor'
REQUIRES_PYTHON = '>=3.7.0'
VERSION = None

# What packages are required for this module to be executed?
REQUIRED = [
    'jsonschema>=3.2.0,<4',
    'pyyaml>=5.3.1,<6'
]

# What packages are optional?
EXTRAS = {
    # 'some-feature': ['requirement'],
}

# Package setup
# -------------

# Import the README and use it as the long description

here = os.path.abspath(os.path.dirname(__file__))

try:
    with io.open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
        long_description = '\n' + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

# Import the LICENSE

with open(os.path.join(here, 'LICENSE')) as f:
    license = f.read()

# Load the package's __version__.py module as a dictionary

about = {}
if not VERSION:
    with open(os.path.join(here, NAME, '__version__.py')) as f:
        exec(f.read(), about)
else:
    about['__version__'] = VERSION

setup(
    name=NAME,
    version=about['__version__'],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/x-rst",
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    project_urls={
        "Documentation": "https://wysdom.readthedocs.io/en/latest/"
    },
    packages=find_packages(exclude=('tests', 'docs')),
    install_requires=REQUIRED,
    extras_require=EXTRAS,
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ]
)



from __future__ import annotations

import json

from ..dom import DOMObject


class ReadsJSON(DOMObject):
    """
    Adds JSON reading and writing functionality to a DOMObject.
    """

    def to_json(self) -> str:
        """
        Serialize the DOM object to JSON.

        return: The DOM object, serialized as a JSON string
        """
        return json.JSONEncoder().encode(self.to_builtin())

    @classmethod
    def from_json(
            cls,
            json_string: str
    ) -> ReadsJSON:
        """
        Create a new DOM object by from a JSON string.

        :param json_string: JSON string to read
        :return:            New DOM object instance
        """
        return cls(
            json.JSONDecoder().decode(json_string)
        )

    @classmethod
    def from_json_file(cls, filename: str) -> ReadsJSON:
        """
        Create a new DOM object from a file on disk.

        :param filename: File path on disk of JSON file
        :return:         New DOM object instance
        """
        with open(filename) as json_file:
            return cls(json.load(json_file))



from typing import Dict, Type, Optional, Any, Tuple

from .SchemaType import SchemaType


class SchemaPrimitive(SchemaType):
    """
    A schema requiring a primitive variable type

    :param python_type: The primitive Python type expected by this schema
    """

    JSON_TYPES: Dict[Type, str] = {
        str: 'string',
        bool: 'boolean',
        int: 'integer',
        float: 'number'
    }

    python_type: Type = None

    def __init__(
            self,
            python_type: Optional[Type] = None
    ) -> None:
        self.python_type = python_type

    def __call__(
            self,
            value: Any,
            dom_info: Tuple = None
    ) -> Any:
        return super().__call__(self.python_type(value))

    @property
    def type_name(self) -> str:
        return self.JSON_TYPES[self.python_type]
